import datetime

import jwt.exceptions as jwt_exceptions
from flask_jwt_extended import JWTManager, exceptions as jwt_extended_exceptions

from bolinette import response, env
from bolinette.exceptions import APIError
from bolinette.services import user_service

jwt = JWTManager()


def init_jwt(app):
    app.config['JWT_TOKEN_LOCATION'] = ['cookies']
    app.config['JWT_COOKIE_SECURE'] = False
    app.config['JWT_ACCESS_COOKIE_PATH'] = '/api/'
    app.config['JWT_REFRESH_COOKIE_PATH'] = '/api/user/token/refresh'
    app.config['JWT_ACCESS_TOKEN_EXPIRES'] = datetime.timedelta(days=1)
    app.config['JWT_REFRESH_TOKEN_EXPIRES'] = datetime.timedelta(days=30)
    app.config['JWT_COOKIE_CSRF_PROTECT'] = True
    app.config['JWT_SECRET_KEY'] = env['JWT_SECRET_KEY']
    jwt.init_app(app)

    @app.errorhandler(jwt_extended_exceptions.JWTExtendedException)
    def handle_auth_error(e):
        return response.unauthorized(str(e))

    @app.errorhandler(jwt_exceptions.PyJWTError)
    def handle_auth_error(e):
        return response.unauthorized(str(e))


@jwt.user_loader_callback_loader
def user_loader_callback(identity):
    try:
        return user_service.get_by_username(identity)
    except APIError:
        return None


@jwt.unauthorized_loader
def unauthorized(_):
    return response.unauthorized('user.unauthorized')


@jwt.expired_token_loader
def expired_token(_):
    return response.unauthorized('user.token.expired')


@jwt.needs_fresh_token_loader
def need_fresh_token():
    return response.unauthorized('user.token.fresh_required')



import inflect
import pydash

from bolinette import console
from bolinette.cli.commands.create_controller import create_controller
from bolinette.cli.commands.create_service import create_service
from bolinette.fs import paths, templating


def create_model(bolinette, **options):
    manifest = paths.read_manifest(bolinette.cwd)
    if manifest is None:
        console.error('No manifest found')
    else:
        module = manifest.get('module')
        path = bolinette.root_path(module)
        origin = bolinette.internal_path('cli', 'files', 'templates')

        model_name = options.get('name')
        class_name = pydash.capitalize(model_name)
        plural_name = inflect.engine().plural(model_name)

        params = {
            'name': model_name,
            'class': class_name,
            'plural': plural_name
        }

        templating.copy(paths.join(origin, 'model.py.jinja2'),
                        paths.join(path, 'models', f'{model_name}.py'), params)
        paths.append(paths.join(path, 'models', '__init__.py'),
                     f'from {module}.models.{model_name} import {class_name}\n')

        if options.get('service', False):
            create_service(bolinette, name=model_name)

        if options.get('controller', False):
            create_controller(bolinette, name=model_name)



from flask import Blueprint

from bolinette.namespaces import Defaults, Route


class Namespace:
    namespaces = []

    def __init__(self, service, url):
        self.service = service
        self.model = service.name
        self.url = url
        self.blueprint = Blueprint(self.model, __name__, url_prefix='/api' + url)

    @staticmethod
    def init_namespaces(app):
        for namespace in Namespace.namespaces:
            app.register_blueprint(namespace)

    def route(self, rule, **options):
        def inner(func):
            endpoint = options.get('endpoint', func.__name__)
            methods = options.get('methods', ['GET'])
            expects = options.get('expects', None)
            returns = options.get('returns', None)
            route_rules = Route(func, self.url, rule, endpoint, methods, expects, returns)
            self.blueprint.add_url_rule(rule, endpoint, route_rules.process, methods=methods)
            return func

        return inner

    def register(self):
        Namespace.namespaces.append(self.blueprint)
    
    @property
    def defaults(self):
        return Defaults(self)



import importlib
import sys

from flask import Flask
from flask_cors import CORS

from bolinette import env, Namespace
from bolinette.database import init_db
from bolinette.jwt import init_jwt
from bolinette.routes import init_routes
from bolinette.documentation import init_docs
from bolinette.fs import paths


class Bolinette:
    def __init__(self, name, **options):
        self.cwd = paths.cwd()
        self.origin = paths.dirname(__file__)
        env_overrides = options.get('env', {})
        self.app = Flask(name, static_url_path='')
        CORS(self.app, supports_credentials=True)
        env.init(self, overrides=env_overrides)
        init_jwt(self.app)
        init_db(self)
        init_routes(self.app)
        init_docs(self.app)
        Namespace.init_namespaces(self.app)

    def instance_path(self, *path):
        return self.root_path('instance', *path)

    def root_path(self, *path):
        return paths.join(self.cwd, *path)

    def internal_path(self, *path):
        return paths.join(self.origin, *path)


def pickup_blnt(cwd):
    manifest = paths.read_manifest(cwd)
    if manifest is not None:
        if cwd not in sys.path:
            sys.path = [cwd] + sys.path
        module = importlib.import_module(manifest.get('module'))
        blnt = getattr(module, 'blnt', None)
        return blnt
    return None



import os
import re

import jinja2

from bolinette.fs import paths


def render(path, params):
    template_path, template_name = paths.split(path)
    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(searchpath=template_path),
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True
    )
    template = jinja_env.get_template(template_name)
    return template.render(**params)


def copy(origin, dest, params):
    content = render(origin, params)
    paths.write(re.sub(r'\.jinja2$', '', dest), content)


def render_directory(origin, dest, params):
    if not os.path.exists(dest):
        paths.mkdir(dest)
    for f in os.listdir(origin):
        if os.path.isdir(paths.join(origin, f)):
            render_directory(paths.join(origin, f), paths.join(dest, f), params)
        if os.path.isfile(paths.join(origin, f)):
            copy(paths.join(origin, f), paths.join(dest, f), params)



from bolinette import transaction, db
from bolinette.database import run_seeders


def init_db(bolinette, **options):
    seed = options.get('seed', False)
    with bolinette.app.app_context():
        db.drop_all()
        db.create_all()
        if seed:
            with transaction:
                run_seeders()



import random
import string

from bolinette import db
from bolinette.marshalling import get_payload, types


def _random_lower(rng, length):
    return ''.join(rng.choices(string.ascii_lowercase, k=length))


def _random_symbols(rng, length):
    return ''.join(rng.choices(string.punctuation, k=length))


def _random_int(rng, a, b):
    return rng.randint(a, b)


def create_mock(m_id, model, key='default'):
    rng = random.Random(hash(f'{model}.{key}.{m_id}'))
    definition = get_payload(f'{model}.{key}')
    mock = {}
    for field in definition.fields:
        if isinstance(field.type, types.classes.String):
            mock[field.name] = _random_lower(rng, 15)
        elif isinstance(field.type, types.classes.Email):
            mock[field.name] = f'{_random_lower(rng, 10)}@{_random_lower(rng, 5)}.com'
        elif isinstance(field.type, types.classes.Password):
            mock[field.name] = (_random_lower(rng, 10) + str(_random_int(rng, 1, 100)) +
                                _random_symbols(rng, 1))
        elif isinstance(field.type, types.classes.Integer):
            mock[field.name] = _random_int(rng, 1, 100)
    return mock


def insert(model, mock):
    entity = model(**mock)
    db.session.add(entity)
    db.session.commit()
    return mock



from flask import after_this_request
from flask_jwt_extended import (
    create_access_token, create_refresh_token, set_access_cookies, set_refresh_cookies,
    get_jwt_identity, jwt_refresh_token_required, unset_jwt_cookies,
    current_user, jwt_required, fresh_jwt_required)

from bolinette import Namespace, response
from bolinette.exceptions import EntityNotFoundError
from bolinette.services import user_service

ns = Namespace(user_service, '/user')


@ns.route('/me',
          methods=['GET'],
          returns={'model': 'user', 'key': 'private'})
@fresh_jwt_required
def me():
    return response.ok('OK', current_user)


@ns.route('/me',
          methods=['PUT'],
          returns={'model': 'user', 'key': 'private'},
          expects={'model': 'user', 'key': 'register', 'patch': True})
@fresh_jwt_required
def update_user(payload):
    user = user_service.get_by_username(get_jwt_identity())
    user = user_service.update(user, payload)
    access_token = create_access_token(identity=user.username, fresh=True)
    refresh_token = create_refresh_token(identity=user.username)

    @after_this_request
    def set_login_cookies(resp):
        set_access_cookies(resp, access_token)
        set_refresh_cookies(resp, refresh_token)
        return resp

    return response.ok('user.updated', user)


@ns.route('/info',
          methods=['GET'],
          returns={'model': 'user', 'key': 'public'})
@jwt_required
def info():
    return response.ok('OK', current_user)


@ns.route('/login',
          methods=['POST'],
          returns={'model': 'user', 'key': 'private'},
          expects={'model': 'user', 'key': 'login'})
def login(payload):
    """
    Logs the user in with the provided credentials
    """
    username = payload['username']
    password = payload['password']
    try:
        user = user_service.get_by_username(username)
    except EntityNotFoundError:
        return response.unauthorized('user.login.wrong_credentials')
    if user is not None:
        if user_service.check_password(user, password):
            access_token = create_access_token(identity=user.username, fresh=True)
            refresh_token = create_refresh_token(identity=user.username)

            @after_this_request
            def set_login_cookies(resp):
                set_access_cookies(resp, access_token)
                set_refresh_cookies(resp, refresh_token)
                return resp

            return response.ok('user.login.success', user)
    return response.unauthorized('user.login.wrong_credentials')


@ns.route('/logout',
          methods=['POST'])
def logout():
    @after_this_request
    def unset_cookies(resp):
        unset_jwt_cookies(resp)
        return resp

    return response.ok('user.logout.success')


@ns.route('/register',
          methods=['POST'],
          returns={'model': 'user', 'key': 'private'},
          expects={'model': 'user', 'key': 'register'})
def register(payload):
    user = user_service.create(payload)
    return response.created('User successfully registered', user)


@ns.route('/token/refresh',
          methods=['POST'])
@jwt_refresh_token_required
def refresh():
    identity = get_jwt_identity()
    access_token = create_access_token(identity=identity, fresh=False)

    @after_this_request
    def reset_cookies(resp):
        set_access_cookies(resp, access_token)
        return resp

    return response.ok('user.token.refreshed')


ns.register()



from bolinette.cli.nodes.node import Node
from bolinette.cli.nodes.command import Command
from bolinette.cli.nodes.argument import Argument
from bolinette.cli.nodes.flag import Flag
from bolinette.cli.nodes.inline import Inline
from bolinette.cli.nodes.ask import Ask



import logging
import sys


class Logger:
    def __new__(cls):
        _logger = logging.getLogger('internal')
        ch = logging.StreamHandler()
        formatter = logging.Formatter('[%(asctime)s] - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        _logger.addHandler(ch)
        return _logger


logger = Logger()


class Console:
    def print(self, text):
        print(text)
    
    def error(self, text):
        print(text, file=sys.stderr)


console = Console()



from bolinette import console
from bolinette.fs import paths, templating


def create_controller(bolinette, **options):
    manifest = paths.read_manifest(bolinette.cwd)
    if manifest is None:
        console.error('No manifest found')
    else:
        module = manifest.get('module')
        path = bolinette.root_path(module)
        origin = bolinette.internal_path('cli', 'files', 'templates')

        model_name = options.get('name')

        params = {
            'module': module,
            'name': model_name
        }

        templating.copy(paths.join(origin, 'controller.py.jinja2'),
                        paths.join(path, 'controllers', f'{model_name}.py'), params)
        paths.append(paths.join(path, 'controllers', '__init__.py'),
                     f'from {module}.controllers.{model_name} import ns as {model_name}_namespace\n')



from bolinette.namespaces.serializing import serializers
from bolinette.namespaces.route import Route
from bolinette.namespaces.defaults import Defaults
from bolinette.namespaces.namespace import Namespace



from bolinette import response


class Defaults:
    def __init__(self, namespace):
        self.model = namespace.model
        self.service = namespace.service
        self.route = namespace.route
    
    def get_all(self, returns='default'):
        @self.route('',
                    methods=['GET'],
                    endpoint=f'get_all_{self.model}',
                    returns={'model': self.model, 'key': returns, 'as_list': True})
        def route():
            return response.ok('OK', self.service.get_all())
    
    def get_one(self, returns='default'):
        @self.route('/<id>',
                    methods=['GET'],
                    endpoint=f'get_one_{self.model}',
                    returns={'model': self.model, 'key': returns})
        def route(**params):
            m_id = params.get('id')
            return response.ok('OK', self.service.get(m_id))
    
    def create(self, returns='default', expects='default'):
        @self.route('',
                    methods=['POST'],
                    endpoint=f'create_{self.model}',
                    returns={'model': self.model, 'key': returns},
                    expects={'model': self.model, 'key': expects})
        def route(**params):
            payload = params.get('payload')
            return response.created(f'{self.model}.created', self.service.create(payload))
    
    def update(self, returns='default', expects='default'):
        @self.route('/<id>',
                    methods=['PUT'],
                    endpoint=f'update_{self.model}',
                    returns={'model': self.model, 'key': returns},
                    expects={'model': self.model, 'key': expects})
        def route(**params):
            m_id = params.get('id')
            payload = params.get('payload')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.updated', self.service.update(entity, payload))
    
    def patch(self, returns='default', expects='default'):
        @self.route('/<id>',
                    methods=['PATCH'],
                    endpoint=f'patch_{self.model}',
                    returns={'model': self.model, 'key': returns},
                    expects={'model': self.model, 'key': expects, 'patch': True})
        def route(**params):
            m_id = params.get('id')
            payload = params.get('payload')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.updated', self.service.patch(entity, payload))
        
    def delete(self, returns='default'):
        @self.route('/<id>',
                    methods=['DELETE'],
                    endpoint=f'delete_{self.model}',
                    returns={'model': self.model, 'key': returns})
        def route(**params):
            m_id = params.get('id')
            entity = self.service.get(m_id)
            return response.ok(f'{self.model}.deleted', self.service.delete(entity))



import os
import random
import string

import yaml


def cwd():
    return os.getcwd()


def random_string(length):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


def mkdir(path):
    os.makedirs(path)


def rename(path, new_path):
    os.rename(path, new_path)


def join(*args):
    return os.path.join(*args)


def dirname(path):
    return os.path.dirname(os.path.realpath(path))


def split(path):
    return os.path.split(path)


def write(path, content, mode='w+'):
    with open(path, mode=mode) as file:
        file.write(content)


def append(path, content):
    write(path, content, mode='a+')


def read_manifest(path):
    try:
        with open(join(path, 'manifest.blnt.yml')) as f:
            return yaml.safe_load(f)
    except FileNotFoundError:
        return None



class Flag:
    def __init__(self, name, flag, desc):
        self.name = name
        self.flag = flag
        self.value = False
        self.desc = desc

    def __repr__(self):
        return f'<Flag {self.name}>'



import importlib

import yaml

from bolinette import pickup_blnt, Bolinette
from bolinette.cli import Loader
from bolinette.cli.nodes import Node, Command
from bolinette.fs import paths


class Parser:
    def __init__(self):
        self.bolinette = pickup_blnt(paths.cwd()) or Bolinette(__name__, env={'SECRET_KEY': 'blnt_cli'})
        assert self.bolinette is not None
        self.nodes = self.parse_nodes()

    def parse_nodes(self):
        with open(self.bolinette.internal_path('cli', 'nodes.yml')) as f:
            return Loader.load_nodes(yaml.safe_load(f))

    def execute(self, argv, cur_node=None, index=1):
        if index >= len(argv):
            return self.print_node_doc(argv[1:index], cur_node)
        nodes = self.nodes if cur_node is None else cur_node.children
        current = None
        for node in nodes:
            if node.name == argv[index]:
                current = node
        if current is not None:
            if isinstance(current, Node):
                return self.execute(argv, current, index + 1)
            elif isinstance(current, Command):
                self.parse_command(argv, current, index + 1)
                if self.check_command(current):
                    self.ask_params(current)
                    self.run_command(current)
                return
        else:
            print('Invalid command: ' + ' '.join(argv[1:index + 1]))
            return self.print_node_doc(argv[1:index], cur_node)

    def print_node_doc(self, argv, node):
        nodes = self.nodes
        if node is not None:
            nodes = node.children
        if len(argv) > 0:
            print('Command: ' + ' '.join(argv))
            print('  Sub commands:')
        else:
            print('Commands:')
        for node in nodes:
            if hasattr(node, 'hidden') and node.hidden:
                continue
            if node.desc is not None:
                print(f'    - {node.name}: {node.desc}')
            else:
                print(f'    - {node.name}')

    def check_command(self, command):
        valid = True
        for inline in command.inline:
            if inline.value is None:
                text = f'Missing parameter: {inline.name}'
                if inline.desc is not None:
                    text += f' ({inline.desc})'
                print(text)
                valid = False
        return valid

    def ask_params(self, command):
        for ask in [a for a in command.ask if a.value is None]:
            text = ask.name
            if ask.desc is not None:
                text += f' ({ask.desc})'
            prompt = '$ '
            if ask.required:
                prompt = '<required> ' + prompt
            if ask.default is not None:
                prompt = f'<default: {ask.default}> ' + prompt
            while True:
                print(text)
                value = input(prompt)
                if not value and ask.default is not None:
                    value = ask.default
                if not not value:
                    break
            if not not value:
                ask.value = value

    def parse_command(self, argv, command, index):
        skip = False
        inline_index = 0
        while index < len(argv):
            current = argv[index]
            if current.startswith('--'):
                for arg in command.args + command.ask:
                    if current == '--' + arg.name:
                        if index + 1 < len(argv):
                            arg.value = argv[index + 1]
                            skip = True
                        else:
                            arg.missing = True
                for flag in command.flags:
                    if current == '--' + flag.name:
                        flag.value = True
            elif current.startswith('-'):
                for arg in command.args:
                    if current == '-' + arg.flag:
                        if index + 1 < len(argv):
                            arg.value = argv[index + 1]
                            skip = True
                        else:
                            arg.missing = True
                for flag in command.flags:
                    if current == '-' + flag.flag:
                        flag.value = True
            else:
                if inline_index < len(command.inline):
                    command.inline[inline_index].value = current
                    inline_index += 1
            index += 1
            if skip:
                index += 1
                skip = False

    def run_command(self, command):
        module = importlib.import_module('bolinette.cli.commands.' + command.command)
        func = getattr(module, command.command)
        params = {'bolinette': self.bolinette}
        for param in command.inline + command.ask + command.flags + command.args:
            params[param.name] = param.value
        for param in command.params:
            params[param] = command.params[param]
        func(**params)



from bolinette import bcrypt
from bolinette.models import User
from bolinette.services import BaseService


class UserService(BaseService):
    def __init__(self):
        super().__init__(User, 'user')

    def get_by_username(self, username):
        return self.get_first_by('username', username)

    def get_by_email(self, email):
        return self.get_first_by('email', email)

    def check_password(self, user, password):
        return bcrypt.check_password_hash(user.password, password)

    def create(self, params):
        params['password'] = bcrypt.generate_password_hash(params['password'])
        return super().create(params)

    def update(self, entity, params):
        if 'password' in params:
            params['password'] = bcrypt.generate_password_hash(params['password'])
        return super().update(entity, params)


user_service = UserService()



from bolinette import response, marshalling
from bolinette.exceptions import EntityNotFoundError

_registered_models = {}
_registered_responses = {}
_registered_payloads = {}


def _get_def(collection, key):
    d = collection.get(key)
    if d is None:
        response.abort(*response.internal_server_error(f'marshalling.unknown_definition:{key}'))
    return d


def get_response(key):
    return _get_def(_registered_responses, key)


def get_payload(key):
    return _get_def(_registered_payloads, key)


def get_model(name):
    return _registered_models.get(name)


def register(model, name):
    def create_defs(collection, params):
        for key, payload in params:
            definition = marshalling.Definition(name, name, key)
            for field in payload:
                definition.fields.append(field)
            collection[definition.key] = definition
    if hasattr(model, 'payloads'):
        create_defs(_registered_payloads, model.payloads())
    if hasattr(model, 'responses'):
        create_defs(_registered_responses, model.responses())
    if name not in _registered_models:
        _registered_models[name] = model


def marshall(definition, entity, skip_none=False, as_list=False):
    if as_list:
        return [marshall(definition, e, skip_none, False) for e in entity]
    data = {}
    for field in definition.fields:
        if isinstance(field, marshalling.Field):
            if field.function is not None:
                value = field.function(entity)
            else:
                value = getattr(entity, field.name, None)
            if field.formatting is not None:
                value = field.formatting(value)
            if not skip_none or value is not None:
                data[field.name] = value
        elif isinstance(field, marshalling.Definition):
            d = get_response(field.key)
            data[field.name] = marshall(d, getattr(entity, field.name), skip_none, False)
        elif isinstance(field, marshalling.List):
            d = get_response(field.element.key)
            data[field.name] = marshall(d, getattr(entity, field.name), skip_none, True)
    return data


def link_foreign_entities(definition, params):
    errors = []
    for field in definition.fields:
        if isinstance(field.type, marshalling.types.classes.ForeignKey):
            value = params.get(field.name, None)
            model = get_model(field.type.model)
            if value is not None and model is not None:
                entity = model.query.filter_by(**{field.type.key: value}).first()
                if entity is None:
                    errors.append((field.type.model, field.type.key, value))
    if len(errors) > 0:
        raise EntityNotFoundError(params=errors)



from bolinette.exceptions import ParamMissingError, ParamConflictError


class Validator:
    def model(self, model, params, **kwargs):
        excluding = kwargs.get('excluding', [])
        errors = []
        valid = {}
        for column in model.__table__.columns:
            key = column.key
            if column.primary_key or key in excluding:
                continue
            value = params.get(key, None)
            if column.unique and value is not None:
                criteria = getattr(model, key) == value
                if model.query.filter(criteria).first() is not None:
                    errors.append((key, value))
            valid[key] = value
        if len(errors) > 0:
            raise ParamConflictError(params=errors)
        return valid

    def payload(self, definition, params, patch=False):
        errors = []
        valid = {}
        for field in definition.fields:
            if patch and field.name not in params:
                continue
            value = params.get(field.name, None)
            if field.required and (value is None or len(str(value)) <= 0):
                errors.append(field.name)
            valid[field.name] = value
        if len(errors) > 0:
            raise ParamMissingError(params=errors)
        return valid


validate = Validator()



class Inline:
    def __init__(self, name, default, desc):
        self.name = name
        self.default = default
        self.value = default
        self.desc = desc

    def __repr__(self):
        return f'<Inline {self.name}>'



import os
from configparser import ConfigParser

from bolinette import logger


class Environment:
    @staticmethod
    def keys(bolinette):
        return {
            'APP_NAME': 'DEFAULT_NAME',
            'APP_DESC': 'DEFAULT_DESCRIPTION',
            'APP_VERSION': '0.0.1',
            'DBMS': 'SQLITE',
            'DEBUG': 'True',
            'FLASK_HOST': '127.0.0.1',
            'FLASK_PORT': '5000',
            'JWT_SECRET_KEY': '',
            'SECRET_KEY': '',
            'WEBAPP_FOLDER': bolinette.root_path('webapp', 'dist')
        }
    
    @staticmethod
    def read_env_file(f):
        return '[DEFAULT]\n' + f.read()

    def __init__(self):
        self.env = {}

    def init(self, bolinette, **options):
        exec_env = options.get('exec_env', 'development')
        overrides = options.get('overrides', {})
        self.env = Environment.keys(bolinette)
        self.env['ENV'] = exec_env
        env_stack = [
            overrides,
            self.load_from_os(),
            self.load_from_file(bolinette, f'{exec_env}.local.env'),
            self.load_from_file(bolinette, f'{exec_env}.env')
        ]
        self.override_env(env_stack)
        self.set_app_config(bolinette.app)

    def set_app_config(self, app):
        app.config['ENV'] = self.env['ENV']
        debug = self.env.get('DEBUG')
        app.config['DEBUG'] = app.config['ENV'] == 'development' if debug is None \
            else debug == 'True'
        secret_key = self.env.get('SECRET_KEY')
        if secret_key is None or len(secret_key) == 0:
            logger.warning('No secret key set! '
                           'Put this "SECRET_KEY=your_secret_key" '
                           f'in instance/{self.env["ENV"]}.local.env')
        app.secret_key = self.env.get('SECRET_KEY')
        app.static_folder = self.env['WEBAPP_FOLDER']
    
    def load_from_os(self):
        keys = {}
        for key in os.environ:
            if key.startswith('BLNT_'):
                keys[key[5:]] = os.environ[key]
        return keys
    
    def load_from_file(self, bolinette, file_name):
        config = ConfigParser()
        try:
            with open(bolinette.instance_path(file_name), 'r') as f:
                config.read_string(Environment.read_env_file(f))
                if (file_env := config['DEFAULT']) is not None:
                    return file_env
                else:
                    return {}
        except FileNotFoundError:
            return {}

    def override_env(self, stack):
        for key in self.env:
            override = self.parse_sources(key, stack)
            if override is not None:
                self.env[key] = override
    
    def parse_sources(self, key, stack):
        for source in stack:
            if key in source:
                return source[key]
        return None

    def __getitem__(self, key):
        return self.env.get(key, None)

    def __setitem__(self, key, value):
        self.env[key] = value

    def get(self, key, default=None):
        item = self[key]
        return item if item is not None else default


env = Environment()



from functools import wraps

from sqlalchemy.exc import SQLAlchemyError

from bolinette import db, response
from bolinette.exceptions import APIError


class Transaction:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            db.session.rollback()
            message, code = None, None
            if issubclass(exc_type, APIError):
                if exc_val.type == APIError.Type.BAD_REQUEST:
                    message, code = response.bad_request(exc_val.messages)
                if exc_val.type == APIError.Type.CONFLICT:
                    message, code = response.conflict(exc_val.messages)
                if exc_val.type == APIError.Type.NOT_FOUND:
                    message, code = response.not_found(exc_val.messages)
            if message is not None and code is not None:
                response.abort(message, code)
        else:
            try:
                db.session.commit()
            except SQLAlchemyError as err:
                db.session.rollback()
                response.abort(*response.internal_server_error(f'global.internal_error:{err}'))


transaction = Transaction()


def transactional(func):
    @wraps(func)
    def inner(*args, **kwargs):
        with transaction:
            return func(*args, **kwargs)
    return inner



from bolinette import validate, db, mapper
from bolinette.exceptions import EntityNotFoundError

_services = {}


class BaseService:
    def __init__(self, model, name):
        self.model = model
        self.name = name
        _services[model] = self

    def service(self, model):
        return _services.get(model)

    def get(self, identifier):
        entity = self.model.query.get(identifier)
        if entity is None:
            raise EntityNotFoundError(model=self.name, key='id', value=identifier)
        return entity

    def get_by(self, key, value):
        return self.model.query.filter_by(**{key: value}).all()

    def get_first_by(self, key, value):
        entity = self.model.query.filter_by(**{key: value}).first()
        if entity is None:
            raise EntityNotFoundError(model=self.name, key=key, value=value)
        return entity

    def get_by_criteria(self, criteria):
        return self.model.query.filter(criteria).all()

    def get_all(self):
        return self.model.query.all()

    def create(self, params):
        params = validate.model(self.model, params)
        entity = self.model(**params)
        db.session.add(entity)
        return entity

    def update(self, entity, params):
        mapper.update(self.model, entity, params)
        return entity
    
    def patch(self, entity, params):
        mapper.update(self.model, entity, params, patch=True)
        return entity

    def delete(self, entity):
        db.session.delete(entity)
        return entity



from bolinette.marshalling.objects import Field, List, Definition
from bolinette.marshalling.types import types
from bolinette.marshalling.definitions import (
    register, marshall, get_payload, get_response, get_model, link_foreign_entities
)



from bolinette import db, seeder


class Role(db.Model):
    __tablename__ = 'roles'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), unique=True, nullable=False)

    def __repr__(self):
        return f'<Role {self.name}>'


@seeder
def role_seeder():
    db.session.add(Role(name='root'))
    db.session.add(Role(name='admin'))



from bolinette.logger import logger, console
from bolinette.environment import env
from bolinette.bcrypt import bcrypt
from bolinette.database import db, seeder
from bolinette.response import response
from bolinette.transaction import transaction, transactional
from bolinette.validator import validate
from bolinette.documentation import docs
from bolinette.namespaces import Namespace
from bolinette.mapper import mapper
from bolinette.jwt import jwt
import bolinette.controllers
from bolinette.bolinette import Bolinette, pickup_blnt


version = '0.1.2'



from flask import request, Response

from bolinette import transaction, marshalling, validate, docs
from bolinette.namespaces import serializers


class Route:
    def __init__(self, func, base_url, url, endpoint, methods, expects, returns):
        self.func = func
        self.base_url = base_url
        self.url = url
        self.endpoint = endpoint
        self.methods = methods
        self.expects = expects
        self.returns = returns
        docs.add_route(self)
    
    def serialize(self, response):
        mime = request.headers.get('Accept', 'application/json')
        serializer = serializers.get(mime) or serializers.default
        return serializer.serialize(response), serializer.mime
    
    def process(self, *args, **kwargs):
        with transaction:
            if self.expects is not None:
                payload = request.get_json(silent=True) or {}
                def_key = f'{self.expects["model"]}.{self.expects["key"]}'
                exp_def = marshalling.get_payload(def_key)
                kwargs['payload'] = validate.payload(
                    exp_def, payload, self.expects.get('patch', False))
                marshalling.link_foreign_entities(exp_def, kwargs['payload'])
            res, code = self.func(*args, **kwargs)
        if self.returns is not None:
            def_key = f'{self.returns["model"]}.{self.returns["key"]}'
            ret_def = marshalling.get_response(def_key)
            if res.get('data') is not None:
                res['data'] = marshalling.marshall(
                    ret_def, res['data'], self.returns.get('skip_none', False),
                    self.returns.get('as_list', False))
        res, mime = self.serialize(res)
        return Response(res, code, mimetype=mime)



class Command:
    def __init__(self, name, command, desc, params, hidden):
        self.name = name
        self.command = command
        self.inline = []
        self.ask = []
        self.args = []
        self.flags = []
        self.desc = desc
        self.params = params
        self.hidden = hidden

    def __repr__(self):
        return f'<Command {self.name}>'



from setuptools import setup, find_packages

import bolinette


def project_packages(module):
    return [m for m in find_packages() if m.startswith(module)]


setup(
    name='Bolinette',
    packages=project_packages('bolinette'),
    include_package_data=True,
    version=bolinette.version,
    license='MIT',
    description='Bolinette, a web framework built on top of Flask',
    author='Pierre Chat',
    author_email='pierrechat89@hotmail.fr',
    url='https://github.com/TheCaptainCat/bolinette',
    keywords=['Flask', 'Bolinette', 'Web', 'Framework'],
    install_requires=[
        'dicttoxml==1.7.4',
        'Flask==1.1.1',
        'Flask-Bcrypt==0.7.1',
        'Flask-Cors==3.0.8',
        'Flask-JWT-Extended==3.24.1',
        'Flask-Script==2.0.6',
        'Flask-SQLAlchemy==2.4.1',
        'inflect==4.0.0',
        'Jinja2==2.10.3',
        'pydash==4.7.6',
        'PyJWT==1.7.1',
        'pytest==5.3.2',
        'PyYAML==5.2',
        'SQLAlchemy==1.3.12',
        'twine==3.1.1',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
    setup_requires=[
        'wheel'
    ],
    entry_points={
        'console_scripts': [
            'blnt=bolinette.cli:main'
        ]
    },
)



from enum import Enum


class ErrorType(Enum):
    NOT_FOUND = 0
    CONFLICT = 1
    BAD_REQUEST = 2


class APIError(Exception):
    Type = ErrorType

    def __init__(self, message, e_type):
        super().__init__(message)
        self.type = e_type


class EntityNotFoundError(APIError):
    def __init__(self, **kwargs):
        super().__init__('EntityNotFoundError', ErrorType.NOT_FOUND)
        params = kwargs.get('params', None)
        model = kwargs.get('model', None)
        key = kwargs.get('key', None)
        value = kwargs.get('value', None)
        if params is None:
            params = [(model, key, value)]
        self.messages = [f'{m}.not_found:{k}:{v}' for m, k, v in params]


class ParamMissingError(APIError):
    def __init__(self, **kwargs):
        super().__init__('ParamMissingError', ErrorType.BAD_REQUEST)
        params = kwargs.get('params', None)
        key = kwargs.get('key', None)
        if params is None:
            params = [key]
        self.messages = [f'param.required:{k}' for k in params]


class ParamConflictError(APIError):
    def __init__(self, **kwargs):
        super().__init__('ParamConflictError', ErrorType.CONFLICT)
        params = kwargs.get('params', None)
        key = kwargs.get('key', None)
        value = kwargs.get('value', None)
        if params is None:
            params = [(key, value)]
        self.messages = [f'param.conflict:{k}:{v}' for k, v in params]



from bolinette import console, version
from bolinette.fs import paths, templating


def create_api(bolinette, **options):
    manifest = paths.read_manifest(bolinette.cwd)
    origin = bolinette.internal_path('cli', 'files')
    if manifest is not None:
        console.error('Manifest file found, it seems Bolinette has already been initialized!')
    else:
        api_name = options.get('name')
        api_desc = options.get('desc')
        api_module = options.get('module')
        params = {
            'secret_key': paths.random_string(64),
            'jwt_secret_key': paths.random_string(64),
            'module': api_module,
            'name': api_name,
            'desc': api_desc,
            'blnt_version': version
        }
        templating.render_directory(paths.join(origin, 'api'), bolinette.cwd, params)
        paths.rename(paths.join(bolinette.cwd, 'server'), paths.join(bolinette.cwd, api_module))



from bolinette.cli.nodes import Node, Command, Inline, Ask, Argument, Flag


class Loader:
    @staticmethod
    def load_nodes(nodes):
        _nodes = []
        for node in nodes:
            _nodes.append(Loader.load_node(node))
        return _nodes

    @staticmethod
    def load_node(node):
        _type = node['type']
        if _type == 'node':
            n = Node(node['name'], node.get('desc'))
            n.children = Loader.load_nodes(node['children'])
            return n
        if _type == 'command':
            return Loader.load_command(node)

    @staticmethod
    def load_command(command):
        c = Command(command['name'], command['command'], command.get('desc'),
                    command.get('params', {}), command.get('hidden', False))
        if 'args' in command:
            c.inline = [Loader.load_inline(a) for a in command['args'] if a['type'] == 'inline']
            c.ask = [Loader.load_ask(a) for a in command['args'] if a['type'] == 'ask']
            c.args = [Loader.load_argument(a) for a in command['args'] if a['type'] == 'arg']
            c.flags = [Loader.load_flag(a) for a in command['args'] if a['type'] == 'flag']
        return c

    @staticmethod
    def load_inline(inline):
        return Inline(inline['name'], inline.get('default'), inline.get('desc'))

    @staticmethod
    def load_ask(ask):
        return Ask(ask['name'], ask.get('required', False), ask.get('default'), ask.get('desc'))

    @staticmethod
    def load_argument(arg):
        return Argument(arg['name'], arg['flag'], arg.get('default'), arg.get('desc'))

    @staticmethod
    def load_flag(flag):
        return Flag(flag['name'], flag['flag'], flag.get('desc'))



from bolinette import db, marshalling

users_roles = db.Table('users_roles',
                       db.Column('u_id', db.Integer, db.ForeignKey('users.id'), primary_key=True),
                       db.Column('r_id', db.Integer, db.ForeignKey('roles.id'), primary_key=True))


class User(db.Model):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(255), unique=True, nullable=False)
    password = db.Column(db.String(255), nullable=False)
    email = db.Column(db.String(255), unique=True, nullable=False)

    roles = db.relationship('Role', secondary=users_roles, lazy='subquery',
                            backref=db.backref('pages', lazy=True))

    @staticmethod
    def payloads():
        yield 'register', [
            marshalling.Field(marshalling.types.string, 'username', required=True),
            marshalling.Field(marshalling.types.email, 'email', required=True),
            marshalling.Field(marshalling.types.password, 'password', required=True)
        ]
        yield 'login', [
            marshalling.Field(marshalling.types.string, 'username', required=True),
            marshalling.Field(marshalling.types.password, 'password', required=True)
        ]

    @staticmethod
    def responses():
        yield 'private', [
            marshalling.Field(marshalling.types.string, 'username'),
            marshalling.Field(marshalling.types.email, 'email')
        ]
        yield 'public', [
            marshalling.Field(marshalling.types.string, 'username')
        ]

    def __repr__(self):
        return f'<User {self.username}>'


marshalling.register(User, 'user')
